@@ -27,7 +27,7 @@
 /* Get size_t, wchar_t and NULL from <stddef.h>.  */
 #define		__need_size_t
 #ifndef __need_malloc_and_calloc
-#if 0
+#ifdef __UCLIBC_HAS_WCHAR__
 # define	__need_wchar_t
 #endif
 # define	__need_NULL
@@ -130,14 +130,16 @@ __extension__ typedef struct
 #define	EXIT_SUCCESS	0	/* Successful exit status.  */
 
 
-#if 0
+#ifdef __UCLIBC_HAS_WCHAR__
 /* Maximum length of a multibyte character in the current locale.  */
-#define	MB_CUR_MAX	(__ctype_get_mb_cur_max ())
-extern size_t __ctype_get_mb_cur_max (void) __THROW;
+#define	MB_CUR_MAX	(_stdlib_mb_cur_max ())
+extern size_t _stdlib_mb_cur_max (void) __THROW;
 #endif
 
+#ifdef __UCLIBC_HAS_FLOATS__
 /* Convert a string to a floating-point number.  */
 extern double atof (__const char *__nptr) __THROW __attribute_pure__;
+#endif /* __UCLIBC_HAS_FLOATS__ */
 /* Convert a string to an integer.  */
 extern int atoi (__const char *__nptr) __THROW __attribute_pure__;
 /* Convert a string to a long integer.  */
@@ -149,6 +151,7 @@ __extension__ extern long long int atoll (__const char *__nptr)
      __THROW __attribute_pure__;
 #endif
 
+#ifdef __UCLIBC_HAS_FLOATS__
 /* Convert a string to a floating-point number.  */
 extern double strtod (__const char *__restrict __nptr,
 		      char **__restrict __endptr) __THROW;
@@ -161,6 +164,7 @@ extern float strtof (__const char *__restrict __nptr,
 extern long double strtold (__const char *__restrict __nptr,
 			    char **__restrict __endptr) __THROW;
 #endif
+#endif /* __UCLIBC_HAS_FLOATS__ */
 
 /* Convert a string to a long integer.  */
 extern long int strtol (__const char *__restrict __nptr,
@@ -198,7 +202,7 @@ extern unsigned long long int strtoull (__const char *__restrict __nptr,
 
 
 #if 0
-/*#ifdef __USE_GNU*/
+#ifdef __USE_GNU
 /* The concept of one static locale per category is not very well
    thought out.  Many applications will need to process its data using
    information from several different locales.  Another application is
@@ -246,11 +250,14 @@ extern long double __strtold_l (__const char *__restrict __nptr,
 				char **__restrict __endptr,
 				__locale_t __loc) __THROW;
 #endif /* GNU */
+#endif /* 0 */
 
 
+#if 0
 /* The internal entry points for `strtoX' take an extra flag argument
    saying whether or not to parse locale-dependent number grouping.  */
 
+#ifdef __UCLIBC_HAS_FLOATS__
 extern double __strtod_internal (__const char *__restrict __nptr,
 				 char **__restrict __endptr, int __group)
      __THROW;
@@ -260,6 +267,7 @@ extern float __strtof_internal (__const char *__restrict __nptr,
 extern long double __strtold_internal (__const char *__restrict __nptr,
 				       char **__restrict __endptr,
 				       int __group) __THROW;
+#endif /* __UCLIBC_HAS_FLOATS__ */
 #ifndef __strtol_internal_defined
 extern long int __strtol_internal (__const char *__restrict __nptr,
 				   char **__restrict __endptr,
@@ -290,8 +298,10 @@ extern unsigned long long int __strtoull_internal (__const char *
 #  define __strtoull_internal_defined	1
 # endif
 #endif /* GCC */
+#endif /* 0 */
 
 #ifdef __USE_EXTERN_INLINES
+#if 0
 /* Define inline functions which call the internal entry points.  */
 
 extern __inline double
@@ -354,6 +364,7 @@ strtoull (__const char * __restrict __nptr, char **__restrict __endptr,
   return __strtoull_internal (__nptr, __endptr, __base, 0);
 }
 # endif
+#endif /* 0 */
 
 extern __inline double
 atof (__const char *__nptr) __THROW
@@ -460,9 +471,11 @@ extern int rand_r (unsigned int *__seed) __THROW;
 #if defined __USE_SVID || defined __USE_XOPEN
 /* System V style 48-bit random number generator functions.  */
 
+#ifdef __UCLIBC_HAS_FLOATS__
 /* Return non-negative, double-precision floating-point value in [0.0,1.0).  */
 extern double drand48 (void) __THROW;
 extern double erand48 (unsigned short int __xsubi[3]) __THROW;
+#endif /* __UCLIBC_HAS_FLOATS__ */
 
 /* Return non-negative, long integer in [0,2^31).  */
 extern long int lrand48 (void) __THROW;
@@ -490,12 +503,14 @@ struct drand48_data
     unsigned long long int __a;	/* Factor in congruential formula.  */
   };
 
+#ifdef __UCLIBC_HAS_FLOATS__
 /* Return non-negative, double-precision floating-point value in [0.0,1.0).  */
 extern int drand48_r (struct drand48_data *__restrict __buffer,
 		      double *__restrict __result) __THROW;
 extern int erand48_r (unsigned short int __xsubi[3],
 		      struct drand48_data *__restrict __buffer,
 		      double *__restrict __result) __THROW;
+#endif /* __UCLIBC_HAS_FLOATS__ */
 
 /* Return non-negative, long integer in [0,2^31).  */
 extern int lrand48_r (struct drand48_data *__restrict __buffer,
@@ -532,6 +547,20 @@ extern void *malloc (size_t __size) __THROW __attribute_malloc__;
 /* Allocate NMEMB elements of SIZE bytes each, all initialized to 0.  */
 extern void *calloc (size_t __nmemb, size_t __size)
      __THROW __attribute_malloc__;
+#if 0
+/* Cope with autoconf's broken AC_FUNC_MALLOC macro, which
+ * redefines malloc to rpl_malloc if it does not detect glibc
+ * style returning-a-valid-pointer-for-malloc(0) behavior.  This
+ * calls malloc() as usual, but if __size is zero, we allocate and
+ * return a 1-byte block instead....  sigh... */ 
+static __inline void *rpl_malloc (size_t __size)
+{
+    if (__size == 0) {
+	__size++; 
+    }
+    return malloc(__size);
+}   
+#endif
 #endif
 
 #ifndef __need_malloc_and_calloc
@@ -569,7 +598,12 @@ extern void abort (void) __THROW __attribute__ ((__noreturn__));
 extern int atexit (void (*__func) (void)) __THROW;
 
 /* The following is used by uClibc in atexit.c and sysconf.c */
-#define __UCLIBC_MAX_ATEXIT     20
+/* We have no limit when __UCLIBC_DYNAMIC_ATEXIT__ is enabled.  */
+#ifdef __UCLIBC_DYNAMIC_ATEXIT__
+# define __UCLIBC_MAX_ATEXIT     INT_MAX
+#else
+# define __UCLIBC_MAX_ATEXIT     20
+#endif
 
 
 #ifdef	__USE_MISC
@@ -727,6 +761,8 @@ __extension__ extern lldiv_t lldiv (long long int __numer,
 #endif
 
 
+#if 0
+#ifdef __UCLIBC_HAS_FLOATS__
 #if defined __USE_SVID || defined __USE_XOPEN_EXTENDED
 /* Convert floating point numbers to strings.  The returned values are
    valid only until another call to the same function.  */
@@ -775,8 +811,10 @@ extern int qfcvt_r (long double __value, int __ndigit,
 		    char *__restrict __buf, size_t __len) __THROW;
 # endif	/* misc */
 #endif	/* use MISC || use X/Open Unix */
+#endif /* __UCLIBC_HAS_FLOATS__ */
+#endif
 
-#if 0
+#ifdef __UCLIBC_HAS_WCHAR__
 /* Return the length of the multibyte character
    in S, which is no longer than N.  */
 extern int mblen (__const char *__s, size_t __n) __THROW;
@@ -796,7 +834,7 @@ extern size_t mbstowcs (wchar_t *__restrict  __pwcs,
 extern size_t wcstombs (char *__restrict __s,
 			__const wchar_t *__restrict __pwcs, size_t __n)
      __THROW;
-#endif
+#endif /* def __UCLIBC_HAS_WCHAR__ */
 
 #ifdef __USE_SVID
 /* Determine whether the string value of RESPONSE matches the affirmation
@@ -850,12 +888,13 @@ extern int unlockpt (int __fd) __THROW;
 extern char *ptsname (int __fd) __THROW;
 #endif
 
-#if 0
-/*#ifdef __USE_GNU*/
+#ifdef __USE_GNU
 /* Store at most BUFLEN characters of the pathname of the slave pseudo
    terminal associated with the master FD is open on in BUF.
    Return 0 on success, otherwise an error number.  */
+#if 0
 extern int ptsname_r (int __fd, char *__buf, size_t __buflen) __THROW;
+#endif
 
 /* Open a master pseudo terminal and return its file descriptor.  */
 extern int getpt (void) __THROW;
