@@ -77,11 +77,22 @@ weak_alias (__libc_write, __write)
 #endif
 
 //#define __NR_open             5
-#ifdef L___libc_open
+#ifdef L___syscall_open
+#define __NR___syscall_open __NR_open
 #include <stdarg.h>
-/* Do not include fcntl.h, so gcc doesn't whine the prototype */
-#define __NR___libc_open __NR_open
-_syscall3(int, __libc_open, const char *, fn, int, flags, mode_t, mode);
+#include <fcntl.h>
+static inline
+_syscall3(int, __syscall_open, const char *, fn, int, flags, __kernel_mode_t, mode);
+int __libc_open (const char * fn, int flags, ...)
+{
+	va_list ap;
+	mode_t mode;
+	va_start(ap, flags);
+	mode = va_arg(ap, mode_t);
+	va_end(ap);
+	return __syscall_open(fn, flags, mode);
+}
+
 weak_alias(__libc_open, open)
 #endif
 
@@ -97,10 +108,15 @@ weak_alias(__libc_close, close)
 // Implemented using wait4 
 
 //#define __NR_creat            8
-#ifdef L_creat
+#ifdef L___syscall_creat
 #include <fcntl.h>
 #ifdef __NR_creat
-_syscall2(int, creat, const char *, file, mode_t, mode);
+#define __NR___syscall_creat __NR_creat
+_syscall2(int, __syscall_creat, const char *, file, __kernel_mode_t, mode);
+int creat (const char *file, mode_t mode)
+{
+	  return __syscall_creat (file, mode);
+}
 #else
 extern int __libc_open (const char *file, int flags, mode_t mode);
 int creat (const char *file, mode_t mode)
@@ -156,15 +172,29 @@ time_t time (time_t *t)
 #endif
 
 //#define __NR_mknod            14
-#ifdef L_mknod
+#ifdef L___syscall_mknod
+#define __NR___syscall_mknod __NR_mknod
 #include <sys/stat.h>
-_syscall3(int, mknod, const char *, path, mode_t, mode, dev_t, dev);
+_syscall3(int, __syscall_mknod, const char *, path, __kernel_mode_t, mode, __kernel_dev_t, dev);
+int mknod(const char *path, mode_t mode, dev_t dev)
+{ 
+	__kernel_dev_t k_dev;
+	/* We must convert the dev_t value to a __kernel_dev_t */
+	k_dev = ((major(dev) & 0xff) << 8) | (minor(dev) & 0xff);
+	return __syscall_mknod(path, mode, k_dev);
+}
 #endif
 
 //#define __NR_chmod            15
-#ifdef L_chmod
+#ifdef L___syscall_chmod
 #include <sys/stat.h>
-_syscall2(int, chmod, const char *, path, mode_t, mode);
+#define __NR___syscall_chmod __NR_chmod
+static inline 
+_syscall2(int, __syscall_chmod, const char *, path, __kernel_mode_t, mode);
+int chmod(const char *path, mode_t mode)
+{
+	return __syscall_chmod(path, mode);
+}
 #endif
 
 /* Old kernels don't have lchown -- do chown instead.  This
@@ -175,9 +205,15 @@ _syscall2(int, chmod, const char *, path, mode_t, mode);
 #endif
 
 //#define __NR_lchown           16
-#ifdef L_lchown
+#ifdef L___syscall_lchown
 #include <unistd.h>
-_syscall3(int, lchown, const char *, path, uid_t, owner, gid_t, group);
+#define __NR___syscall_lchown __NR_lchown
+static inline 
+_syscall3(int, __syscall_lchown, const char *, path, __kernel_uid_t, owner, __kernel_gid_t, group);
+int lchown(const char *path, uid_t owner, gid_t group)
+{
+	return __syscall_lchown(path, owner, group);
+}
 #endif
 
 
@@ -220,7 +256,20 @@ _syscall1(int, umount, const char *, specialfile);
 #endif
 
 //#define __NR_setuid           23
-// See setuid.c
+#ifdef L___syscall_setuid
+#define __NR___syscall_setuid __NR_setuid
+#include <unistd.h>
+static inline 
+_syscall1(int, __syscall_setuid, __kernel_uid_t, uid);
+int setuid(uid_t uid)
+{
+	if (uid == (uid_t) ~0 || uid != (uid_t) ((int) uid)) {
+		__set_errno (EINVAL);
+		return -1;
+	}
+	return(__syscall_setuid(uid));
+}
+#endif
 
 //#define __NR_getuid           24
 #ifdef L_getuid
@@ -253,7 +302,7 @@ int stime(const time_t *when)
 #ifdef L___ptrace
 #include <sys/ptrace.h>
 #define __NR___ptrace __NR_ptrace
-_syscall4(long, __ptrace, enum __ptrace_request, request, pid_t, pid,
+_syscall4(long, __ptrace, enum __ptrace_request, request, __kernel_pid_t, pid,
 		void*, addr, void*, data);
 #endif
 
@@ -401,10 +450,16 @@ int nice (int incr)
 //See sync.c
 
 //#define __NR_kill             37
-#ifdef L_kill
+#ifdef L___syscall_kill
 #include <signal.h>
 #undef kill
-_syscall2(int, kill, pid_t, pid, int, sig);
+#define __NR___syscall_kill __NR_kill
+static inline
+_syscall2(int, __syscall_kill, __kernel_pid_t, pid, int, sig);
+int kill(pid_t pid, int sig)
+{
+	return(__syscall_kill(pid, sig));
+}
 #endif
 
 //#define __NR_rename           38
@@ -414,9 +469,15 @@ _syscall2(int, rename, const char *, oldpath, const char *, newpath);
 #endif
 
 //#define __NR_mkdir            39
-#ifdef L_mkdir
+#ifdef L___syscall_mkdir
 #include <sys/stat.h>
-_syscall2(int, mkdir, const char *, pathname, mode_t, mode);
+#define __NR___syscall_mkdir __NR_mkdir
+static inline 
+_syscall2(int, __syscall_mkdir, const char *, pathname, __kernel_mode_t, mode);
+int mkdir(const char * pathname, mode_t mode)
+{
+	return(__syscall_mkdir(pathname, mode));
+}
 #endif
 
 //#define __NR_rmdir            40
@@ -448,9 +509,15 @@ _syscall1(clock_t, times, struct tms *, buf);
 //#define __NR_brk              45
 
 //#define __NR_setgid           46
-#ifdef L_setgid
+#ifdef L___syscall_setgid
 #include <unistd.h>
-_syscall1(int, setgid, gid_t, gid);
+#define __NR___syscall_setgid __NR_setgid
+static inline 
+_syscall1(int, __syscall_setgid, __kernel_gid_t, gid);
+int setgid(gid_t gid)
+{
+	return(__syscall_setgid(gid));
+}
 #endif
 
 //#define __NR_getgid           47
@@ -513,25 +580,16 @@ _syscall1(int, acct, const char *, filename);
 //#define __NR_lock             53
 
 //#define __NR_ioctl            54
-#ifdef L__ioctl
+#ifdef L___syscall_ioctl
 #include <stdarg.h>
 #include <sys/ioctl.h>
-#define __NR__ioctl __NR_ioctl
-extern int _ioctl(int fd, int request, void *arg);
-
-_syscall3(int, _ioctl, int, fd, int, request, void *, arg);
-
-int ioctl(int fd, unsigned long int request, ...)
-{
-	void *arg;
-	va_list list;
-
-	va_start(list, request);
-	arg = va_arg(list, void *);
-
-	va_end(list);
-	return _ioctl(fd, request, arg);
-}
+#define __NR___syscall_ioctl __NR_ioctl
+extern int __syscall_ioctl(int fd, int request, void *arg);
+_syscall3(int, __syscall_ioctl, int, fd, int, request, void *, arg);
+#if !defined (__powerpc__)
+#include "ioctl.c"
+/* Also see ioctl.c and powerpc/ioctl.c */
+#endif
 #endif
 
 //#define __NR_fcntl            55
@@ -560,9 +618,15 @@ weak_alias(__libc_fcntl, fcntl)
 //#define __NR_mpx              56
 
 //#define __NR_setpgid          57
-#ifdef L_setpgid
+#ifdef L___syscall_setpgid
 #include <unistd.h>
-_syscall2(int, setpgid, pid_t, pid, pid_t, pgid);
+#define __NR___syscall_setpgid __NR_setpgid
+static inline
+_syscall2(int, __syscall_setpgid, __kernel_pid_t, pid, __kernel_pid_t, pgid);
+int setpgid(pid_t pid, pid_t pgid)
+{
+	return(__syscall_setpgid(pid, pgid));
+}
 #endif
 
 //#define __NR_ulimit           58
@@ -570,9 +634,15 @@ _syscall2(int, setpgid, pid_t, pid, pid_t, pgid);
 //#define __NR_oldolduname      59
 
 //#define __NR_umask            60
-#ifdef L_umask
+#ifdef L___syscall_umask
 #include <sys/stat.h>
-_syscall1(mode_t, umask, mode_t, mask);
+#define __NR___syscall_umask __NR_umask
+static inline 
+_syscall1(__kernel_mode_t, __syscall_umask, __kernel_mode_t, mask);
+mode_t umask(mode_t mask)
+{
+	return(__syscall_umask(mask));
+}
 #endif
 
 //#define __NR_chroot           61
@@ -616,11 +686,11 @@ _syscall0(pid_t, setsid);
 
 //#define __NR_sigaction        67
 #ifndef __NR_rt_sigaction
-#define __NR___sigaction __NR_sigaction
-#ifdef L___sigaction
+#ifdef L___syscall_sigaction
+#define __NR___syscall_sigaction __NR_sigaction
 #include <signal.h>
 #undef sigaction
-_syscall3(int, __sigaction, int, signum, const struct sigaction *, act,
+_syscall3(int, __syscall_sigaction, int, signum, const struct sigaction *, act,
 		  struct sigaction *, oldact);
 #endif
 #endif
@@ -630,15 +700,27 @@ _syscall3(int, __sigaction, int, signum, const struct sigaction *, act,
 //#define __NR_ssetmask         69
 
 //#define __NR_setreuid         70
-#ifdef L_setreuid
+#ifdef L___syscall_setreuid
 #include <unistd.h>
-_syscall2(int, setreuid, uid_t, ruid, uid_t, euid);
+#define __NR___syscall_setreuid __NR_setreuid
+static inline
+_syscall2(int, __syscall_setreuid, __kernel_uid_t, ruid, __kernel_uid_t, euid);
+int setreuid(uid_t ruid, uid_t euid)
+{
+	return(__syscall_setreuid(ruid, euid));
+}
 #endif
 
 //#define __NR_setregid         71
-#ifdef L_setregid
+#ifdef L___syscall_setregid
 #include <unistd.h>
-_syscall2(int, setregid, gid_t, rgid, gid_t, egid);
+#define __NR___syscall_setregid __NR_setregid
+static inline
+_syscall2(int, __syscall_setregid, __kernel_gid_t, rgid, __kernel_gid_t, egid);
+int setregid(gid_t rgid, gid_t egid)
+{
+	return(__syscall_setregid(rgid, egid));
+}
 #endif
 
 //#define __NR_sigsuspend       72
@@ -749,16 +831,56 @@ _syscall2(int, settimeofday, const struct timeval *, tv,
 #endif
 
 //#define __NR_getgroups        80
-#ifdef L_getgroups
+#ifdef L___syscall_getgroups
 #include <unistd.h>
-_syscall2(int, getgroups, int, size, gid_t *, list);
+#define __NR___syscall_getgroups __NR_getgroups
+static inline
+_syscall2(int, __syscall_getgroups, int, size, __kernel_gid_t *, list);
+#define MIN(a,b) (((a)<(b))?(a):(b))
+int getgroups(int n, gid_t *groups)
+{
+	if (unlikely(n < 0)) {
+		__set_errno(EINVAL);
+		return -1;
+	} else {
+		int i, ngids;
+		__kernel_gid_t kernel_groups[n = MIN(n, sysconf(_SC_NGROUPS_MAX))];
+		ngids = __syscall_getgroups(n, kernel_groups);
+		if (n != 0 && ngids > 0) {
+			for (i = 0; i < ngids; i++) {
+				(groups)[i] = kernel_groups[i];
+			}
+		}
+		return ngids;
+	}
+}
 #endif
 
 //#define __NR_setgroups        81
-#ifdef L_setgroups
+#ifdef L___syscall_setgroups
 #include <unistd.h>
 #include <grp.h>
-_syscall2(int, setgroups, size_t, size, const gid_t *, list);
+#define __NR___syscall_setgroups __NR_setgroups
+static inline
+_syscall2(int, __syscall_setgroups, size_t, size, const __kernel_gid_t *, list);
+int setgroups (size_t n, const gid_t *groups)
+{
+	if (n > (size_t)sysconf(_SC_NGROUPS_MAX)) {
+		__set_errno (EINVAL);
+		return -1;
+	} else {
+		size_t i;
+		__kernel_gid_t kernel_groups[n];
+		for (i = 0; i < n; i++) {
+			kernel_groups[i] = (groups)[i];
+			if (groups[i] != (gid_t) ((__kernel_gid_t)groups[i])) {
+				__set_errno (EINVAL);
+				return -1;
+			}
+		}
+		return(__syscall_setgroups(n, kernel_groups));
+	}
+}
 #endif
 
 //#define __NR_select           82
@@ -858,21 +980,46 @@ _syscall2(int, ftruncate, int, fd, __off_t, length);
 #endif
 
 //#define __NR_fchmod           94
-#ifdef L_fchmod
+#ifdef L___syscall_fchmod
 #include <sys/stat.h>
-_syscall2(int, fchmod, int, fildes, mode_t, mode);
+#define __NR___syscall_fchmod __NR_fchmod
+static inline 
+_syscall2(int, __syscall_fchmod, int, fildes, __kernel_mode_t, mode);
+int fchmod(int fildes, mode_t mode)
+{
+	return(__syscall_fchmod(fildes, mode));
+}
 #endif
 
 //#define __NR_fchown           95
-#ifdef L_fchown
+#ifdef L___syscall_fchown
 #include <unistd.h>
-_syscall3(int, fchown, int, fd, uid_t, owner, gid_t, group);
+#define __NR___syscall_fchown __NR_fchown
+static inline
+_syscall3(int, __syscall_fchown, int, fd, __kernel_uid_t, owner, __kernel_gid_t, group);
+int fchown(int fd, uid_t owner, gid_t group)
+{
+	return(__syscall_fchown(fd, owner, group));
+}
 #endif
 
 //#define __NR_getpriority      96
-#ifdef L_getpriority
+#ifdef L___syscall_getpriority
 #include <sys/resource.h>
-_syscall2(int, getpriority, __priority_which_t, which, id_t, who);
+#define __NR___syscall_getpriority __NR_getpriority
+_syscall2(int, __syscall_getpriority, __priority_which_t, which, id_t, who);
+/* The return value of __syscall_getpriority is biased by this value
+ * to avoid returning negative values.  */
+#define PZERO 20
+int getpriority (enum __priority_which which, id_t who)
+{
+	int res;
+
+	res = __syscall_getpriority(which, who);
+	if (res >= 0)
+		res = PZERO - res;
+	return res;
+}
 #endif
 
 //#define __NR_setpriority      97
@@ -945,31 +1092,72 @@ _syscall2(int, getitimer, __itimer_which_t, which, struct itimerval *, value);
 #endif
 
 //#define __NR_stat             106
-#ifdef L_stat
+#ifdef L___syscall_stat
+#define __NR___syscall_stat __NR_stat
 #include <unistd.h>
-#include <sys/stat.h>
-_syscall2(int, stat, const char *, file_name, struct stat *, buf);
-#if ! defined __NR_stat64 && defined __UCLIBC_HAVE_LFS__
+#define _SYS_STAT_H
+#include <bits/stat.h>
+#include "xstatconv.h"
+_syscall2(int, __syscall_stat, const char *, file_name, struct kernel_stat *, buf);
+int stat(const char * file_name, struct stat * buf)
+{
+	int result;
+	struct kernel_stat kbuf;
+	result = __syscall_stat(file_name, &kbuf);
+	if (result == 0) {
+		__xstat_conv(&kbuf, buf);
+	}
+	return result;
+}
+#if ! defined __NR_stat64 && defined __UCLIBC_HAS_LFS__
 weak_alias(stat, stat64);
 #endif
 #endif
 
 //#define __NR_lstat            107
-#ifdef L_lstat
+#ifdef L___syscall_lstat
+#define __NR___syscall_lstat __NR_lstat
 #include <unistd.h>
-#include <sys/stat.h>
-_syscall2(int, lstat, const char *, file_name, struct stat *, buf);
-#if ! defined __NR_lstat64 && defined __UCLIBC_HAVE_LFS__
+#define _SYS_STAT_H
+#include <bits/stat.h>
+#include <bits/kernel_stat.h>
+#include "xstatconv.h"
+_syscall2(int, __syscall_lstat, const char *, file_name, struct kernel_stat *, buf);
+int lstat(const char * file_name, struct stat * buf)
+{
+	int result;
+	struct kernel_stat kbuf;
+	result = __syscall_lstat(file_name, &kbuf);
+	if (result == 0) {
+		__xstat_conv(&kbuf, buf);
+	}
+	return result;
+}
+#if ! defined __NR_lstat64 && defined __UCLIBC_HAS_LFS__
 weak_alias(lstat, lstat64);
 #endif
 #endif
 
 //#define __NR_fstat            108
-#ifdef L_fstat
+#ifdef L___syscall_fstat
+#define __NR___syscall_fstat __NR_fstat
 #include <unistd.h>
-#include <sys/stat.h>
-_syscall2(int, fstat, int, filedes, struct stat *, buf);
-#if ! defined __NR_fstat64 && defined __UCLIBC_HAVE_LFS__
+#define _SYS_STAT_H
+#include <bits/stat.h>
+#include <bits/kernel_stat.h>
+#include "xstatconv.h"
+_syscall2(int, __syscall_fstat, int, fd, struct kernel_stat *, buf);
+int fstat(int fd, struct stat * buf)
+{
+	int result;
+	struct kernel_stat kbuf;
+	result = __syscall_fstat(fd, &kbuf);
+	if (result == 0) {
+		__xstat_conv(&kbuf, buf);
+	}
+	return result;
+}
+#if ! defined __NR_fstat64 && defined __UCLIBC_HAS_LFS__
 weak_alias(fstat, fstat64);
 #endif
 #endif
@@ -978,8 +1166,10 @@ weak_alias(fstat, fstat64);
 
 //#define __NR_iopl             110
 #ifdef L_iopl
+/* For arm there is a totally different implementation */
+#if !defined(__arm__)
 /* Tuns out the m68k unistd.h kernel header is broken */
-#	if defined __UCLIBC_HAS_MMU__ && defined __NR_iopl && ! defined(__mc68000__)
+#	if defined __UCLIBC_HAS_MMU__ && defined __NR_iopl && ( !defined(__mc68000__))
 		_syscall1(int, iopl, int, level);
 #	else
 		int iopl(int level)
@@ -988,6 +1178,7 @@ weak_alias(fstat, fstat64);
 			return -1;
 		}
 #	endif
+# endif
 #endif
 
 //#define __NR_vhangup          111
@@ -1002,8 +1193,14 @@ _syscall0(int, vhangup);
 //#define __NR_vm86old          113
 
 //#define __NR_wait4            114
-#ifdef L_wait4
-_syscall4(int, wait4, pid_t, pid, int *, status, int, opts, void *, rusage);
+#ifdef L___syscall_wait4
+#define __NR___syscall_wait4 __NR_wait4
+static inline
+_syscall4(int, __syscall_wait4, __kernel_pid_t, pid, int *, status, int, opts, void *, rusage);
+int wait4(pid_t pid, int * status, int opts, void * rusage)
+{
+	return(__syscall_wait4(pid, status, opts, rusage));
+}
 #endif
 
 //#define __NR_swapoff          115
@@ -1053,6 +1250,12 @@ _syscall1(int, uname, struct utsname *, buf);
 #endif
 
 //#define __NR_modify_ldt       123
+#ifdef __NR_modify_ldt
+#ifdef L_modify_ldt
+_syscall3(int, modify_ldt, int, func, void *, ptr, unsigned long, bytecount);
+weak_alias(modify_ldt, __modify_ldt);
+#endif
+#endif
 
 //#define __NR_adjtimex         124
 #ifdef L_adjtimex
@@ -1117,8 +1320,14 @@ _syscall4(int, quotactl, int, cmd, const char *, special , int, id, caddr_t, add
 #endif
 
 //#define __NR_getpgid          132
-#ifdef L_getpgid
-_syscall1(pid_t, getpgid, pid_t, pid);
+#ifdef L___syscall_getpgid
+#define __NR___syscall_getpgid __NR_getpgid
+static inline
+_syscall1(__kernel_pid_t, __syscall_getpgid, __kernel_pid_t, pid);
+pid_t getpgid(pid_t pid)
+{
+	return(__syscall_getpgid(pid));
+}
 #endif
 
 //#define __NR_fchdir           133
@@ -1141,17 +1350,29 @@ _syscall2(int, bdflush, int, __func, long int, __data);
 
 //#define __NR_setfsuid         138
 #ifdef __NR_setfsuid
-#ifdef L_setfsuid
+#ifdef L___syscall_setfsuid
 #include <sys/fsuid.h>
-_syscall1(int, setfsuid, uid_t, uid);
+#define __NR___syscall_setfsuid __NR_setfsuid
+static inline
+_syscall1(int, __syscall_setfsuid, __kernel_uid_t, uid);
+int setfsuid(uid_t uid)
+{
+	return(__syscall_setfsuid(uid));
+}
 #endif
 #endif
 
 //#define __NR_setfsgid         139
 #ifdef __NR_setfsgid
-#ifdef L_setfsgid
+#ifdef L___syscall_setfsgid
 #include <sys/fsuid.h>
-_syscall1(int, setfsgid, gid_t, gid);
+#define __NR___syscall_setfsgid __NR_setfsgid
+static inline
+_syscall1(int, __syscall_setfsgid, __kernel_gid_t, gid);
+int setfsgid(gid_t gid)
+{
+	return(__syscall_setfsgid(gid));
+}
 #endif
 #endif
 
@@ -1159,11 +1380,7 @@ _syscall1(int, setfsgid, gid_t, gid);
 //See llseek.c
 
 //#define __NR_getdents         141
-#ifdef L_getdents
-#include <unistd.h>
-#include <dirent.h>
-_syscall3(int, getdents, int, fd, char *, dirp, size_t, count);
-#endif
+// See getdents.c
 
 //#define __NR__newselect       142
 #ifdef L__newselect
@@ -1208,9 +1425,15 @@ _syscall3(ssize_t, writev, int, filedes, const struct iovec *, vector, int,
 #endif
 
 //#define __NR_getsid           147
-#ifdef L_getsid
+#ifdef L___syscall_getsid
 #include <unistd.h>
-_syscall1(pid_t, getsid, pid_t, pid);
+#define __NR___syscall_getsid __NR_getsid
+static inline
+_syscall1(__kernel_pid_t, __syscall_getsid, __kernel_pid_t, pid);
+pid_t getsid(pid_t pid)
+{
+	return(__syscall_getsid(pid));
+}
 #endif
 
 //#define __NR_fdatasync        148
@@ -1222,6 +1445,27 @@ _syscall1(int, fdatasync, int, fd);
 #endif
 
 //#define __NR__sysctl          149
+#ifdef __NR__sysctl
+#ifdef L__sysctl
+#include <linux/sysctl.h>
+_syscall1(int, _sysctl, struct __sysctl_args *, args);
+int sysctl(int *name, int nlen, void *oldval, size_t *oldlenp,
+			void *newval, size_t newlen)
+{
+	struct __sysctl_args args =
+	{
+		name: name,
+		nlen: nlen,
+		oldval: oldval,
+		oldlenp: oldlenp,
+		newval: newval,
+		newlen: newlen
+	};
+
+	return _sysctl(&args);
+}
+#endif
+#endif
 
 //#define __NR_mlock            150
 #ifdef L_mlock
@@ -1257,33 +1501,57 @@ _syscall1(int, fdatasync, int, fd);
 
 //#define __NR_sched_setparam   154
 #ifdef __NR_sched_setparam
-#ifdef L_sched_setparam
+#ifdef L___syscall_sched_setparam
 #include <sched.h>
-_syscall2(int, sched_setparam, pid_t, pid, const struct sched_param *, p);
+#define __NR___syscall_sched_setparam __NR_sched_setparam
+static inline
+_syscall2(int, __syscall_sched_setparam, __kernel_pid_t, pid, const struct sched_param *, p);
+int sched_setparam(pid_t pid, const struct sched_param * p)
+{
+	return(__syscall_sched_setparam(pid, p));
+}
 #endif
 #endif
 
 //#define __NR_sched_getparam   155
 #ifdef __NR_sched_getparam
-#ifdef L_sched_getparam
+#ifdef L___syscall_sched_getparam
 #include <sched.h>
-_syscall2(int, sched_getparam, pid_t, pid, struct sched_param *, p);
+#define __NR___syscall_sched_getparam __NR_sched_getparam
+static inline
+_syscall2(int, __syscall_sched_getparam, __kernel_pid_t, pid, struct sched_param *, p);
+int sched_getparam(pid_t pid, struct sched_param * p)
+{
+	return(__syscall_sched_getparam(pid, p));
+}
 #endif
 #endif
 
 //#define __NR_sched_setscheduler       156
 #ifdef __NR_sched_setscheduler
-#ifdef L_sched_setscheduler
+#ifdef L___syscall_sched_setscheduler
 #include <sched.h>
-_syscall3(int, sched_setscheduler, pid_t, pid, int, policy, const struct sched_param *, p);
+#define __NR___syscall_sched_setscheduler __NR_sched_setscheduler
+static inline
+_syscall3(int, __syscall_sched_setscheduler, __kernel_pid_t, pid, int, policy, const struct sched_param *, p);
+int sched_setscheduler(pid_t pid, int policy, const struct sched_param * p)
+{
+	return(__syscall_sched_setscheduler(pid, policy, p));
+}
 #endif
 #endif
 
 //#define __NR_sched_getscheduler       157
 #ifdef __NR_sched_getscheduler
-#ifdef L_sched_getscheduler
+#ifdef L___syscall_sched_getscheduler
 #include <sched.h>
-_syscall1(int, sched_getscheduler, pid_t, pid);
+#define __NR___syscall_sched_getscheduler __NR_sched_getscheduler
+static inline
+_syscall1(int, __syscall_sched_getscheduler, __kernel_pid_t, pid);
+int sched_getscheduler(pid_t pid)
+{
+	return(__syscall_sched_getscheduler(pid));
+}
 #endif
 #endif
 
@@ -1313,9 +1581,15 @@ _syscall1(int, sched_get_priority_min, int, policy);
 
 //#define __NR_sched_rr_get_interval    161
 #ifdef __NR_sched_rr_get_interval
-#ifdef L_sched_rr_get_interval
+#ifdef L___syscall_sched_rr_get_interval
 #include <sched.h>
-_syscall2(int, sched_rr_get_interval, pid_t, pid, struct timespec *, tp);
+#define __NR___syscall_sched_rr_get_interval __NR_sched_rr_get_interval
+static inline
+_syscall2(int, __syscall_sched_rr_get_interval, __kernel_pid_t, pid, struct timespec *, tp);
+int sched_rr_get_interval(pid_t pid, struct timespec * tp)
+{
+	return(__syscall_sched_rr_get_interval(pid, tp));
+}
 #endif
 #endif
 
@@ -1335,7 +1609,39 @@ _syscall4(__ptr_t, mremap, __ptr_t, old_address, size_t, old_size, size_t, new_s
 #endif
 
 //#define __NR_setresuid                164
+#ifdef __NR_setresuid
+#ifdef L___syscall_setresuid
+#define __NR___syscall_setresuid __NR_setresuid
+static inline
+_syscall3(int, __syscall_setresuid, __kernel_uid_t, rgid, __kernel_uid_t, egid, __kernel_uid_t, sgid);
+int setresuid(uid_t ruid, uid_t euid, uid_t suid)
+{
+	return(__syscall_setresuid(ruid, euid, suid));
+}
+#endif
+#endif
+
 //#define __NR_getresuid                165
+#ifdef __NR_getresuid
+#ifdef L___syscall_getresuid
+#define __NR___syscall_getresuid __NR_getresuid
+static inline
+_syscall3(int, __syscall_getresuid, __kernel_uid_t *, ruid, __kernel_uid_t *, euid, __kernel_uid_t *, suid);
+int getresuid (uid_t *ruid, uid_t *euid, uid_t *suid)
+{
+	int result;
+	__kernel_uid_t k_ruid, k_euid, k_suid;
+	result = __syscall_getresuid(&k_ruid, &k_euid, &k_suid);
+	if (result == 0) {
+		*ruid = (uid_t) k_ruid;
+		*euid = (uid_t) k_euid;
+		*suid = (uid_t) k_suid;
+	}
+	return result;
+}
+#endif
+#endif
+
 //#define __NR_vm86                     166
 
 //#define __NR_query_module             167
@@ -1363,17 +1669,49 @@ _syscall3(int, poll, struct pollfd *, fds, unsigned long int, nfds, int, timeout
 //nfsservctl	EXTRA	nfsservctl	i:ipp	nfsservctl
 
 //#define __NR_setresgid                170
+#ifdef __NR_setresgid
+#ifdef L___syscall_setresgid
+#define __NR___syscall_setresgid __NR_setresgid
+static inline
+_syscall3(int, __syscall_setresgid, __kernel_gid_t, rgid, __kernel_gid_t, egid, __kernel_gid_t, sgid);
+int setresgid(gid_t rgid, gid_t egid, gid_t sgid)
+{
+	return(__syscall_setresgid(rgid, egid, sgid));
+}
+#endif
+#endif
+
 //#define __NR_getresgid                171
+#ifdef __NR_getresgid
+#ifdef L___syscall_getresgid
+#define __NR___syscall_getresgid __NR_getresgid
+static inline
+_syscall3(int, __syscall_getresgid, __kernel_gid_t *, egid, __kernel_gid_t *, rgid, __kernel_gid_t *, sgid);
+int getresgid(gid_t *rgid, gid_t *egid, gid_t *sgid)
+{
+	int result;
+	__kernel_gid_t k_rgid, k_egid, k_sgid;
+	result = __syscall_getresgid(&k_rgid, &k_egid, &k_sgid);
+	if (result == 0) { 
+		*rgid = (gid_t) k_rgid;
+		*egid = (gid_t) k_egid;
+		*sgid = (gid_t) k_sgid;
+	}
+	return result;
+}
+#endif
+#endif
+
 //#define __NR_prctl                    172
 
 //#define __NR_rt_sigreturn             173
 //#define __NR_rt_sigaction             174
 #ifdef __NR_rt_sigaction
-#define __NR___rt_sigaction __NR_rt_sigaction
-#ifdef L___rt_sigaction
+#define __NR___syscall_rt_sigaction __NR_rt_sigaction
+#ifdef L___syscall_rt_sigaction
 #include <signal.h>
 #undef sigaction
-_syscall4(int, __rt_sigaction, int, signum, const struct sigaction *, act, 
+_syscall4(int, __syscall_rt_sigaction, int, signum, const struct sigaction *, act, 
 		struct sigaction *, oldact, size_t, size); 
 #endif
 #endif
@@ -1473,9 +1811,15 @@ int sigsuspend (const sigset_t *mask)
 // See pread_write.c
 
 //#define __NR_chown                    182
-#ifdef L_chown
+#ifdef L___syscall_chown
 #include <unistd.h>
-_syscall3(int, chown, const char *, path, uid_t, owner, gid_t, group);
+#define __NR___syscall_chown __NR_chown
+static inline
+_syscall3(int, __syscall_chown, const char *, path, __kernel_uid_t, owner, __kernel_gid_t, group);
+int chown(const char * path, uid_t owner, gid_t group)
+{
+	return(__syscall_chown(path, owner, group));
+}
 #endif
 
 //#define __NR_getcwd                   183
@@ -1520,7 +1864,7 @@ _syscall2(int, sigaltstack, const struct sigaltstack *, ss, struct sigaltstack *
 #ifdef L_sendfile
 #include <unistd.h>
 #include <sys/sendfile.h>
-_syscall4(ssize_t,sendfile, int, out_fd, int, in_fd, __off_t *, offset, size_t, count)
+_syscall4(ssize_t,sendfile, int, out_fd, int, in_fd, __off_t *, offset, size_t, count);
 #endif
 #endif
 
@@ -1557,30 +1901,69 @@ int getrlimit (__rlimit_resource_t resource, struct rlimit *rlimits)
 
 
 //#define __NR_stat64             195
-#ifdef L_stat64
-#if defined __NR_stat64 && defined __UCLIBC_HAVE_LFS__
+#ifdef L___syscall_stat64
+#if defined __NR_stat64 && defined __UCLIBC_HAS_LFS__
+#define __NR___syscall_stat64 __NR_stat64
 #include <unistd.h>
 #include <sys/stat.h>
-_syscall2(int, stat64, const char *, file_name, struct stat64 *, buf);
-#endif /* __UCLIBC_HAVE_LFS__ */
+#include <bits/kernel_stat.h>
+#include "xstatconv.h"
+_syscall2(int, __syscall_stat64, const char *, file_name, struct kernel_stat64 *, buf);
+int stat64(const char * file_name, struct stat64 * buf)
+{
+	int result;
+	struct kernel_stat64 kbuf;
+	result = __syscall_stat64(file_name, &kbuf);
+	if (result == 0) {
+		__xstat64_conv(&kbuf, buf);
+	}
+	return result;
+}
+#endif /* __UCLIBC_HAS_LFS__ */
 #endif
 
 //#define __NR_lstat64            196
-#ifdef L_lstat64
-#if defined __NR_lstat64 && defined __UCLIBC_HAVE_LFS__
+#ifdef L___syscall_lstat64
+#if defined __NR_lstat64 && defined __UCLIBC_HAS_LFS__
+#define __NR___syscall_lstat64 __NR_lstat64
 #include <unistd.h>
 #include <sys/stat.h>
-_syscall2(int, lstat64, const char *, file_name, struct stat64 *, buf);
-#endif /* __UCLIBC_HAVE_LFS__ */
+#include <bits/kernel_stat.h>
+#include "xstatconv.h"
+_syscall2(int, __syscall_lstat64, const char *, file_name, struct kernel_stat64 *, buf);
+int lstat64(const char * file_name, struct stat64 * buf)
+{
+	int result;
+	struct kernel_stat64 kbuf;
+	result = __syscall_lstat64(file_name, &kbuf);
+	if (result == 0) {
+		__xstat64_conv(&kbuf, buf);
+	}
+	return result;
+}
+#endif /* __UCLIBC_HAS_LFS__ */
 #endif
 
 //#define __NR_fstat64            197
-#ifdef L_fstat64
-#if defined __NR_fstat64 && defined __UCLIBC_HAVE_LFS__
+#ifdef L___syscall_fstat64
+#if defined __NR_fstat64 && defined __UCLIBC_HAS_LFS__
+#define __NR___syscall_fstat64 __NR_fstat64
 #include <unistd.h>
 #include <sys/stat.h>
-_syscall2(int, fstat64, int, filedes, struct stat64 *, buf);
-#endif /* __UCLIBC_HAVE_LFS__ */
+#include <bits/kernel_stat.h>
+#include "xstatconv.h"
+_syscall2(int, __syscall_fstat64, int, filedes, struct kernel_stat64 *, buf);
+int fstat64(int fd, struct stat64 * buf)
+{
+	int result;
+	struct kernel_stat64 kbuf;
+	result = __syscall_fstat64(fd, &kbuf);
+	if (result == 0) {
+		__xstat64_conv(&kbuf, buf);
+	}
+	return result;
+}
+#endif /* __UCLIBC_HAS_LFS__ */
 #endif
 
 
@@ -1606,30 +1989,30 @@ _syscall2(int, fstat64, int, filedes, struct stat64 *, buf);
 //#define __NR_pivot_root		217
 #ifdef __NR_pivot_root
 #ifdef L_pivot_root
-_syscall2(int, pivot_root, const char *, new_root, const char *, put_old)
+_syscall2(int, pivot_root, const char *, new_root, const char *, put_old);
 #endif
 #endif
 
 //#define __NR_mincore		218
 //#define __NR_madvise		219
+#ifdef __NR_madvise
+#ifdef L_madvise
+_syscall3(int, madvise, void*, __addr, size_t, __len, int, __advice);
+#endif
+#endif
+
 //#define __NR_madvise1		219	/* delete when C lib stub is removed */
 
 //#define __NR_getdents64		220
-#ifdef L_getdents64
-#ifdef __UCLIBC_HAVE_LFS__
-#include <unistd.h>
-#include <dirent.h>
-_syscall3(int, getdents64, int, fd, char *, dirp, size_t, count);
-#endif /* __UCLIBC_HAVE_LFS__ */
-#endif
+// See getdents64.c
 
 //#define __NR_fcntl64		221
 #ifdef L__fcntl64
-#ifdef __UCLIBC_HAVE_LFS__
-#ifdef __NR_fcntl64
-#define __NR__fcntl64 __NR_fcntl64
+#ifdef __UCLIBC_HAS_LFS__
 #include <stdarg.h>
 #include <fcntl.h>
+#ifdef __NR_fcntl64
+#define __NR__fcntl64 __NR_fcntl64
 extern int _fcntl64(int fd, int cmd, long arg);
 
 _syscall3(int, _fcntl64, int, fd, int, cmd, long, arg);
@@ -1646,8 +2029,18 @@ int fcntl64(int fd, int command, ...)
 	return _fcntl64(fd, command, arg);
 }
 #else
-extern int __libc_fcntl(int fd, int command, ...);
-weak_alias(__libc_fcntl, fcntl64)
+extern int _fcntl(int fd, int cmd, long arg);
+int fcntl64(int fd, int command, ...)
+{
+	long arg;
+	va_list list;
+
+	va_start(list, command);
+	arg = va_arg(list, long);
+
+	va_end(list);
+	return _fcntl(fd, command, arg);
+}
 #endif
 #endif
 #endif
